/* NO USER CONFIGURABLE PARAMETERS HERE! */

/* Path Integral Monte Carlo code developed by A. Balaz 
   (antun@ipb.ac.rs) for the paper:
         
   "Systematic Speedup in the Convergence of Path Integrals"
   by A. Bogojevic, A. Balaz, and A. Belic
   Phys. Rev. Lett. 94 (2005) 180403

   Address: Institute of Physics, Belgrade, Serbia
   Scientific Computing Laboratory, http://www.scl.rs/speedup/

   Public use and modification of this code is allowed providing 
   the above paper is properly acknowledged. The author would be 
   grateful for all information and/or comments regarding the use   
   of the code.  
*/

/* The function main() reads command line parameters, initializes variables,
   calls the Monte Carlo function mc(), and prints the output.
*/

#include <main.h>

int main(int argc, char **argv)
{
   double a, b; /* Initial and final position. */

   int s, n; /* Maximal number of time steps is 2^s; n is the loop variable that
                runs from 0 (or 1) to s, and corresponds to the current
                number of time steps 2^n. */

   double Nmc, Nmc_proc; /* Nmc is the number of Monte Carlo samples, and Nmc_proc
                            is the number of Monte Carlo samples per process. */

   long seed; /* Initialization seed for the SPRNG random number generator. */

   int npar; /* Number of parameters of the potential enetered by the user. */

   double *amp, *err; /* Arrays of amplitudes and Monte Carlo errors for the
                         numbers of time steps 2^0, ..., 2^s. */

   double *q; /* Array that will hold the generated paths. */

   int i; /* Loop variable used during the initialization of *norm. */

   double *mpi_data; /* Array that will hold all the variables during the MPI communication. */

   double tcpu, tcpu_all; /* Time of execution per process and overall time of execution. */

   struct timeval tv;  /* Structures defined in sys/time.h, used for the calculation of */
   struct timezone tz; /* the execution time. */

   int np, myid, proc; /* MPI variables: np is the number of processes, myid is the id of
                          the current process, and proc is a loop variable running 
                          from 0 to np - 1. */

   MPI_Init(&argc, &argv); /* MPI initialization. */
   MPI_Comm_size(MPI_COMM_WORLD, &np); /* MPI function that returns np. */
   MPI_Comm_rank(MPI_COMM_WORLD, &myid); /* MPI function that returns myid. */

   if((argc < 7) && (myid == 0)) /* Prints the usage if the number of */
   {                             /* command line arguments is < 7. */
      fprintf(stderr, "Usage: %s T a b s Nmc seed par[0] ...\n", argv[0]);
      fprintf(stderr, "This simulation calculates the probability amplitude\n");
      fprintf(stderr, "in imaginary time for the transition from the\n");
      fprintf(stderr, "initial state q(t=0)=a to the final state q(t=T)=b.\n");
      fprintf(stderr, "The number of time steps N will be 2^0, ..., 2^s.\n");
      fprintf(stderr, "Nmc is the number of MC samples\n");
      fprintf(stderr, "seed is used by the SPRNG (random number generator).\n");
      fprintf(stderr, "par[0], ... are additional parameters defining the potential\n");
      fprintf(stderr, "(initialize as many as needed and use them in the same order\n");
      fprintf(stderr, "in the user supplied function V0 in file p.c)\n");
      exit(EXIT_FAILURE);
   }

   T = atof(argv[1]); /* Initialization of variables entered on the command line. */
   a = atof(argv[2]);
   b = atof(argv[3]);
   s = atol(argv[4]);
   Nmc = atof(argv[5]);

   Nmc_proc = Nmc / np; /* Number of Monte Carlo samples per process. */

   seed = atol(argv[6]);

   par = dvector(0, argc - 7); /* Allocation of memory for the parameters of the potential. */

   for(npar = 0; npar < argc - 7; npar ++) par[npar] = atof(argv[7 + npar]); /* Initialization of *par. */

   dpi = 8. * atan(1.); /* Initialization of dpi = 2 pi = 8 * atan(1). */

   amp = dvector(0, s); /* Allocation of memory for amplitudes. */

   err = dvector(0, s); /* Allocation of memory for errors. */

   mpi_data = dvector(0, (2 * (s + 1) + 1) * np); /* Allocation of memory for MPI communication. */

   two = ivector(0, s); /* Allocation of memory for array with powers of 2. */

   norm = dvector(0, s); /* Allocation of memory for norms of discretized path integrals
                            with the number of time steps 2^0, ..., 2^s, used in func(). */

   two[0] = 1; /* Initialization of array with powers of 2. */

   for(n = 1; n <= s; n ++)
   {
      two[n] = 2 * two[n - 1];
   }

   distrexp = dvector(0, two[s]); /* Allocation of memory for exponents of the Gaussian probability
                                     distribution functions, used in func(). */

   q = dvector(0, two[s]); /* Allocation of memory for paths. */

   dsigma2 = dvector(0, s); /* Allocation of memory for array of 2 sigma^2 values. Sigmas are
                               standard deviations of Gaussians used as probability density functions. */

   dsigma2inv = dvector(0, s); /* Allocation of memory for array of 1 / (2 sigma^2) values. */

   for(n = 0; n <= s; n ++) /* Initialization of arrays *amp and *err to zeros. */
   {
      amp[n] = 0;
      err[n] = 0;
   }

   for(n = 0; n <= s; n ++) /* Initialization of array of 2 sigma^2 and 1 / (2 sigma^2) values, */
   {                        /* where sigma^2 = eps = T / 2^n. */
      dsigma2[n] = T / two[n];
      dsigma2inv[n] = 1. / dsigma2[n];
   }

   for(n = 0; n <= s; n ++) /* Initialization of array of the logarithms of norms. Norm is a product */
   {                        /* of (2 pi T / 2^n)^(- 2^n / 2) (normalization of the path integral in
                               the discretization with 2^n time steps) and normalization of Gaussians, i.e.
                               the square root of appropriate 2 pi sigma^2 for each path node. */
      norm[n] = - 0.5 * log(dpi) - 0.5 * two[n] * log(T / two[n]);

      for(i = 1; i <= n; i ++) norm[n] += 0.5 * two[i - 1] * log(0.5 * dsigma2[i]);
   }

   q[0] = a; /* Boundary conditions. */
   q[two[s]] = b;

   distrexp[0] = 0; /* q[0] is not generated, so distrexp[0] = 0. It is introduced for simplicity in distr(). */

   streamnum = myid; /* Initialization of the SPRNG random number generator. */
   nstreams = np;
   stream = init_sprng(SPRNG_CMRG, streamnum, nstreams, seed, SPRNG_DEFAULT);

   if(myid == 0) /* Printing the header of the output. */
   {
      fprintf(stdout, "N amplitude error time[s]\n");
      fflush(stdout);
   }

   gettimeofday(&tv, &tz);               /* The initial time. Saving it here, we will avoid taking */
   tcpu = tv.tv_sec + tv.tv_usec * 1e-6; /* into account the time spent on MPI initialization. */

   mc(Nmc_proc, q, amp, err, s); /* Calling the function that will carry out the Monte Carlo calculation. */

   gettimeofday(&tv, &tz);                      /* Calculates the time spent in mc(), as the */
   tcpu = tv.tv_sec + tv.tv_usec * 1e-6 - tcpu; /* difference between the current and initial time. */

   for(n = 0; n <= s; n ++) /* Initialization of array for MPI communication. */
   {
      mpi_data[2 * n] = amp[n];
      mpi_data[2 * n + 1] = err[n];
   }

   mpi_data[2 * (s + 1)] = tcpu;

   /* MPI communication. Data from all processes are gathered in *mpi_data of process 0. */
   MPI_Gather(mpi_data, 2 * (s + 1) + 1, MPI_DOUBLE, mpi_data, 2 * (s + 1) + 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

   if(myid == 0) /* Processing the data if myid is 0. */
   {
      for(n = 0; n <= s; n ++) /* Initialization of arrays. */
      {
         amp[n] = 0;
         err[n] = 0;
      }

      tcpu_all = 0; /* Initialization of overall execution time. */

      for(proc = 0; proc < np; proc ++) /* Loop over processes. */
      {
         for(n = 0; n <= s; n ++) /* Loop over n. */
         {
            amp[n] += mpi_data[proc * (2 * (s + 1) + 1) + 2 * n]; /* Accumulation of f / p values. */
            err[n] += mpi_data[proc * (2 * (s + 1) + 1) + 2 * n + 1]; /* Accumulation of (f / p)^2 values. */
         }

         tcpu_all += mpi_data[proc * (2 * (s + 1) + 1) + 2 * (s + 1)]; /* Accumulation of tcpu_all. */
      }

      amp[0] /= Nmc; /* Calculation of the amplitude for n = 0, i.e. 2^0 = 1 time step. */
      err[0] = 0; /* For n = 0, the MC error is zero, as there are no integrals to calculate. */

      fprintf(stdout, "%d %1.15le %1.15le\n", two[0], amp[0], err[0]); /* Printing the n = 0 output. */
      fflush(stdout);

      for(n = 1; n <= s; n++) /* Calculation and printing of the amplitudes and errors for n > 0. */
      {
         amp[n] /= Nmc;
         err[n] = (err[n] / Nmc - amp[n] * amp[n]) / (Nmc - 1);

         if(err[n] > 0) err[n] = sqrt(err[n]); /* Avoiding roundoff error. */
         else err[n] = 0;

         fprintf(stdout, "%d %1.15le %1.15le\n", two[n], amp[n], err[n]);
         fflush(stdout);
      }

      /* Printing np, tcpu_all, Nmc. */
      fprintf(stdout, "Overall execution time of %d processes: %le seconds.\n", np, tcpu_all);
      fprintf(stdout, "Number of Monte Carlo samples: %le.\n", Nmc);
      fflush(stdout);
   }

   /* Deallocation of memory. */
   free_sprng(stream);
   free_dvector(dsigma2inv, 0, s);
   free_dvector(dsigma2, 0, s);
   free_dvector(q, 0, two[s]);
   free_dvector(distrexp, 0, two[s]);
   free_dvector(norm, 0, s);
   free_ivector(two, 0, s);
   free_dvector(mpi_data, 0, (2 * (s + 1) + 1) * np);
   free_dvector(err, 0, s);
   free_dvector(amp, 0, s);
   free_dvector(par, 0, argc - 7);

   MPI_Finalize(); /* Exiting MPI. */

   exit(EXIT_SUCCESS);
}

/* The function mc() does the main Monte Carlo loop. In the loop it calls the
   function distr() to generate the path, and enters the loop over the number
   of time steps. In this loop it calls the function func() to calculate
   f / p (for a given n), the quantity we are averaging to obtain the integral
   of the function f. The array *amp accumulates f / p values for each n, while the
   array *err accumulates (f / p)^2 values, needed in order to calculate the error.
*/

void mc(double Nmc, double *q, double *amp, double *err, int s)
{
   double i; /* Loop variable for MC samples. It is of the double type to allow for
                large Nmc values. */
   int n; /* Loop variable for the number of time steps. */
   double foverp; /* f / p value, returned by the func(). */

   for(i = 0; i < Nmc; i ++) /* Loop over the MC samples. */
   {
      distr(q, s); /* Generating the path, with 2^s time steps. */

      for(n = 0; n <= s; n ++) /* Loop over the number of time steps, 2^0, ..., 2^s. */
      {
         foverp = func(q, T, s, n); /* Calculates f / p for a given n. */
         amp[n] += foverp; /* Accumulates f / p for a given n. */
         err[n] += foverp * foverp; /* Accumulates (f / p)^2 for a given n. */
      }
   }

   return;
}

/* Function distr() generates the paths as arrays (q[0], ..., q[2^s]) using
   the bisection algorithm.
*/

void distr(double *q, int s)
{
   int n; /* Loop variable for bisection level. */
   int i; /* Loop variable for path nodes that need to be generated at bisection level n. */
   double temp; /* Temporary variable. */

   for(n = 1; n <= s; n ++) /* Loop over bisection level. */
   {
      for(i = 0; i < two[n - 1]; i ++) /* Loop over path nodes that need to be generated at */
      {                                /* bisection level n. */

         /* Using the Box-Mueller method to generate the random number temp from the Gaussian centered
            at zero, with 2 sigma^2 = dsigma2[n]. */
         temp = sqrt( - dsigma2[n] * log(sprng(stream))) * cos(dpi * sprng(stream));

         /* At bisection level n, the path node is centered at the mid-point of the nearest path nodes
            generated at bisection level n - 1. */
         q[(1 + 2 * i) * two[s - n]] = 0.5 * (q[i * two[s - n + 1]] + q[(1 + i) * two[s - n + 1]]) + temp;

         /* The exponent of the Gaussian probability density function. It will be substracted from the action in
            func(), which should return f / p. The normalization of the Gaussian is taken care of in *norm. */
         distrexp[(1 + 2 * i) * two[s - n]]  = temp * temp * dsigma2inv[n];
      }
   }

   return;
}
